@@ -3,6 +3,7 @@
 #
 
 package Forks::Super::Queue;
+
 use Forks::Super::Config;
 use Forks::Super::Debug qw(:all);
 use Forks::Super::Tie::Enum;
@@ -24,16 +25,16 @@ our $NEXT_DEFERRED_ID = -100000;
 our $OLD_SIG;
 our $VERSION = $Forks::Super::Debug::VERSION;
 our $MAIN_PID = $$;
-our $QUEUE_MONITOR_LAUNCHED = 0;
+#our $QUEUE_MONITOR_LAUNCHED = 0;
 our $_LOCK = 0; # ??? can this prevent crash -- no, but it can cause deadlock
 our $CHECK_FOR_REAP = 1;
+our $QUEUE_DEBUG = $ENV{FORKS_SUPER_QUEUE_DEBUG} || 0;
 # set flag if the program is shutting down. Use flag in queue_job()
 # to suppress warning messages
 our $DURING_GLOBAL_DESTRUCTION = 0;
 
 # use var $Forks::Super::QUEUE_INTERRUPT, not lexical package var
 
-
 sub get_default_priority {
   my $q = $DEFAULT_QUEUE_PRIORITY;
   $DEFAULT_QUEUE_PRIORITY -= 1.0E-6;
@@ -41,11 +42,20 @@ sub get_default_priority {
 }
 
 sub init {
-  $QUEUE_MONITOR_FREQ = 30;
-  tie $Forks::Super::QUEUE_INTERRUPT, 'Forks::Super::Tie::Enum',
-    ('', keys %SIG);
-  $Forks::Super::QUEUE_INTERRUPT = 'USR1' if grep {/USR1/} keys %SIG;
-  $INHIBIT_QUEUE_MONITOR = $^O eq 'MSWin32';
+  tie $QUEUE_MONITOR_FREQ, 
+    'Forks::Super::Queue::QueueMonitorFreq', 30;
+
+  tie $Forks::Super::QUEUE_INTERRUPT, 
+    'Forks::Super::Queue::QueueInterrupt', ('', keys %SIG);
+
+  tie $INHIBIT_QUEUE_MONITOR, 
+    'Forks::Super::Queue::InhibitQueueMonitor', 
+    $^O eq 'MSWin32'; # XXX - or any other $^O with crippled signal framework
+
+  if (grep {/USR1/} keys %SIG) {
+    $Forks::Super::QUEUE_INTERRUPT = 'USR1';
+  }
+
 }
 
 sub init_child {
@@ -75,9 +85,20 @@ sub init_child {
 # $Forks::Super::QUEUE_INTERRUPT signals to this
 #
 sub _launch_queue_monitor {
-  return unless Forks::Super::Config::CONFIG('SIGUSR1');
-  return if defined $QUEUE_MONITOR_PID;
-  return if $QUEUE_MONITOR_LAUNCHED++;
+  if (!Forks::Super::Config::CONFIG('SIGUSR1')) {
+    debug("_lqm returning: no SIGUSR1") if $QUEUE_DEBUG;
+    return;
+  }
+  if (defined $QUEUE_MONITOR_PID) {
+    debug("_lqm returning: \$QUEUE_MONITOR_PID defined") if $QUEUE_DEBUG;
+    return;
+  }
+  if (!(defined $Forks::Super::QUEUE_INTERRUPT
+	&& $Forks::Super::QUEUE_INTERRUPT)) {
+    debug("_lqm returning: \$Forks::Super::QUEUE_INTERRUPT not set")
+      if $QUEUE_DEBUG;
+    return;
+  }
 
   $OLD_SIG = $SIG{$Forks::Super::QUEUE_INTERRUPT};
   $SIG{$Forks::Super::QUEUE_INTERRUPT} = \&Forks::Super::Queue::check_queue;
@@ -91,7 +112,7 @@ sub _launch_queue_monitor {
   }
   if ($QUEUE_MONITOR_PID == 0) {
     $0 = "$QUEUE_MONITOR_PPID:QMon";
-    if ($DEBUG) {
+    if ($DEBUG || $QUEUE_DEBUG) {
       debug("Launching queue monitor process $$ ",
 	    "SIG $Forks::Super::QUEUE_INTERRUPT ",
 	    "PPID $QUEUE_MONITOR_PPID ",
@@ -105,7 +126,7 @@ sub _launch_queue_monitor {
     }
     for (;;) {
       sleep $QUEUE_MONITOR_FREQ;
-      if ($DEBUG) {
+      if ($DEBUG || $QUEUE_DEBUG) {
 	debug("queue monitor $$ passing signal to $QUEUE_MONITOR_PPID");
       }
       CORE::kill $Forks::Super::QUEUE_INTERRUPT, $QUEUE_MONITOR_PPID;
@@ -119,12 +140,12 @@ sub _kill_queue_monitor {
   if (defined $QUEUE_MONITOR_PPID && $$ == $QUEUE_MONITOR_PPID) {
     if (defined $QUEUE_MONITOR_PID && $QUEUE_MONITOR_PID > 0) {
 
-      if ($DEBUG) {
+      if ($DEBUG || $QUEUE_DEBUG) {
 	debug("killing queue monitor $QUEUE_MONITOR_PID");
       }
       CORE::kill 'INT', $QUEUE_MONITOR_PID;
       my $z = CORE::waitpid $QUEUE_MONITOR_PID, 0;
-      if ($DEBUG) {
+      if ($DEBUG || $QUEUE_DEBUG) {
 	debug("kill queue monitor result: $z");
       }
 
@@ -157,7 +178,7 @@ sub queue_job {
     $job->{state} = 'DEFERRED';
     $job->{pid} = $NEXT_DEFERRED_ID--;
     $Forks::Super::ALL_JOBS{$job->{pid}} = $job;
-    if ($DEBUG) {
+    if ($DEBUG || $QUEUE_DEBUG) {
       debug("queueing job ", $job->toString());
     }
   }
@@ -176,7 +197,7 @@ our $_REAP;
 
 sub _check_for_reap {
   if ($CHECK_FOR_REAP && $_REAP > 0) {
-    if ($DEBUG) {
+    if ($DEBUG || $QUEUE_DEBUG) {
       debug("reap during queue examination -- restart");
     }
     return 1;
@@ -205,7 +226,7 @@ sub run_queue {
   #   order by priority
   #   go through the list and attempt to launch each job in order.
 
-  debug('run_queue(): examining deferred jobs') if $DEBUG;
+  debug('run_queue(): examining deferred jobs') if $DEBUG || $QUEUE_DEBUG;
   my $job_was_launched;
   do {
     $job_was_launched = 0;
@@ -258,7 +279,7 @@ sub run_queue {
     }
   } while ($job_was_launched);
 
-  if (0) {   # suspend/resume under development
+  if (0) {   # suspend/resume callback under development
     my @suspended_jobs = grep { $_->{state} eq 'SUSPENDED'
 			      } @Forks::Super::ALL_JOBS;
     my @active_and_suspendable_jobs 
@@ -296,4 +317,96 @@ sub check_queue {
   return;
 }
 
+#############################################################################
+
+# when $Forks::Super::Queue::QUEUE_MONITOR_FREQ is updated,
+# we should restart the queue monitor.
+
+sub Forks::Super::Queue::QueueMonitorFreq::TIESCALAR {
+  my ($class,$value) = @_;
+  $value = int $value;
+  if ($value == 0) {
+    $value = 1;
+  } elsif ($value < 0) {
+    $value = 30;
+  }
+  debug("new F::S::Q::QueueMonitorFreq obj") if $QUEUE_DEBUG;
+  return bless \$value, $class;
+}
+
+sub Forks::Super::Queue::QueueMonitorFreq::FETCH {
+  my $self = shift;
+  debug("F::S::Q::QueueMonitorFreq::FETCH: $$self") if $QUEUE_DEBUG;
+  return $$self;
+}
+
+sub Forks::Super::Queue::QueueMonitorFreq::STORE {
+  my ($self,$new_value) = @_;
+  $new_value = int($new_value) || 1;
+  $new_value = 30 if $new_value < 0;
+  if ($new_value == $$self) {
+    debug("F::S::Q::QueueMonitorFreq::STORE noop $$self") if $QUEUE_DEBUG;
+    return $$self;
+  }
+  if ($QUEUE_DEBUG) {
+    debug("F::S::Q::QueueMonitorFreq::STORE $$self <== $new_value");
+  }
+  $$self = $new_value;
+  _kill_queue_monitor();
+  run_queue();
+  _launch_queue_monitor() if @QUEUE > 0;
+}
+
+#############################################################################
+
+# When $Forks::Super::Queue::INHIBIT_QUEUE_MONITOR is changed to non-zero,
+# always call _kill_queue_monitor.
+
+sub Forks::Super::Queue::InhibitQueueMonitor::TIESCALAR {
+  my ($class,$value) = @_;
+  $value = 0+!!$value;
+  return bless \$value, $class;
+}
+
+sub Forks::Super::Queue::InhibitQueueMonitor::FETCH {
+  my $self = shift;
+  return $$self;
+}
+
+sub Forks::Super::Queue::InhibitQueueMonitor::STORE {
+  my ($self, $new_value) = @_;
+  $new_value = 0+!!$new_value;
+  if ($$self != $new_value && $new_value) {
+    _kill_queue_monitor();
+  }
+  $$self = $new_value;
+  return $$self;
+}
+
+#############################################################################
+
+# Restart queue monitor if value for $QUEUE_INTERRUPT is changed.
+
+*Forks::Super::Queue::QueueInterrupt::TIESCALAR
+  = \&Forks::Super::Tie::Enum::TIESCALAR;
+
+*Forks::Super::Queue::QueueInterrupt::FETCH
+  = \&Forks::Super::Tie::Enum::FETCH;
+
+sub Forks::Super::Queue::QueueInterrupt::STORE {
+  my ($self, $new_value) = @_;
+  if (uc $new_value eq uc Forks::Super::Tie::Enum::_get_value($self)) {
+    return; # no change
+  }
+  if (!Forks::Super::Tie::Enum::_has_attr($self,$new_value)) {
+    return; # invalid assignment
+  }
+  _kill_queue_monitor();
+  $Forks::Super::Tie::Enum::VALUE{$self} = $new_value;
+  _launch_queue_monitor() if @QUEUE > 0;
+  return;
+}
+
+#############################################################################
+
 1;
